package com.xyz.abc.datasource;

import static java.lang.Math.abs;
import static java.lang.Math.sqrt;
import static junit.framework.Assert.assertEquals;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.infinispan.config.Configuration;
import org.infinispan.distribution.ch.ConsistentHash;
import org.infinispan.distribution.ch.ConsistentHashHelper;
import org.infinispan.distribution.ch.DefaultConsistentHash;
import org.infinispan.distribution.ch.ExperimentalDefaultConsistentHash;
import org.infinispan.distribution.ch.TopologyAwareConsistentHash;
import org.infinispan.remoting.transport.Address;
import org.infinispan.remoting.transport.jgroups.JGroupsAddress;
import org.junit.Test;

public class PnlKeyHashTester
{   
	//@Test
	public void testDistributionwithDiffernetNumberOfNodes() 
	{      
		final int numKeys = 500;
		final int[] numNodes = {1, 2, 3, 4, 5};

		for (int nn : numNodes) 
		{         
			List<PnLDataKey> keys = buildKeys(numKeys);      
			final int numVirtualNodes = nn*4;
			doTestDistribution(numKeys, nn, numVirtualNodes, keys);      
		}   
	}   
	
	//@Test
	public void testDistributionWithDifferentNumberOfKeys() 
	{      
		final int[] numKeyArr = {100, 500, 1000, 2000, 5000, 10000, 20000};
		final int numNodes = 3;      
		final int numVirtualNodes = numNodes*4;

		for (int numKeys : numKeyArr) 
		{         
			List<PnLDataKey> keys = buildKeys(numKeys);      
			doTestDistribution(numKeys, numNodes, numVirtualNodes, keys);
		}   
	}   

	@Test
	public void testDistributionWithDifferentNumberOfVirtualNodes() 
	{      
		final int numKeys = 1000;
		final int numNodes = 3;   
		final int[] numVirtualNodesArr = {numNodes*2, numNodes*4, numNodes*8, numNodes*16, numNodes*32};

		for (int numVirtualNodes : numVirtualNodesArr) 
		{         
			List<PnLDataKey> keys = buildKeys(numKeys);      
			doTestDistribution(numKeys, numNodes, numVirtualNodes, keys);
		}   
	}   
	
	private void doTestDistribution(int numKeys, int numNodes, int numVirtualNodes, List<?> keys) 
	{      
		ConsistentHash ch = createConsistentHash(numNodes, numVirtualNodes);
		Map<Address, Integer> distribution = new HashMap<Address, Integer>();      
		for (Object key : keys)
		{         
			Address a = ch.locate(key, 1).get(0);         
			if (distribution.containsKey(a)) 
			{            
				int i = distribution.get(a);            
				distribution.put(a, i + 1);         
			} 
			else 
			{            
				distribution.put(a, 1);         
			}      
		}
		System.out.printf("\nTesting distribution with %d keys, %d nodes, %d virtual nodes\n", numKeys, numNodes, numVirtualNodes);      
		//System.out.println("" + distribution);      
		// calc numbers      
		ArrayList<Integer> counts = new ArrayList<Integer>(distribution.values());      
		Collections.sort(counts);
		assertEquals(numNodes, counts.size());      
		// instead we add a 0 for all the nonexistent keys in the distribution map and do the calculations      
		for (int i = 0; i < numNodes - counts.size(); i++)
			counts.add(0, 0);
		double mean = 0;      
		int sum = 0;      
		for (Integer count : counts) 
			sum += count;
		assertEquals(sum, numKeys);
		mean = sum / numNodes;
		double variance = 0;
		for (Integer count : counts) 
			variance += (count - mean) * (count - mean);      
		double stdDev = sqrt(variance);      
		double avgAbsDev = 0;      
		for (Integer count : counts) 
			avgAbsDev += abs(count - mean);      
		avgAbsDev /= numNodes;      
		int median = counts.get(numNodes / 2);      
		ArrayList<Integer> medianDevs = new ArrayList<Integer>(numNodes);      
		for (Integer count : counts) 
			medianDevs.add(abs(count - median));      
		Collections.sort(medianDevs);      
		int medianAbsDev = medianDevs.get(numNodes / 2);      
		System.out.printf("Mean = %f, median = %d\n", mean, median);      
		System.out.printf("Standard deviation = %.3f, or %.3f%%\n", stdDev, stdDev / mean * 100);      
		System.out.printf("Average absolute deviation = %.3f, or %.3f%%\n", avgAbsDev, avgAbsDev / mean * 100);      
		System.out.printf("Median absolute deviation = %d, or %.3f%%\n", medianAbsDev, (double)medianAbsDev / mean * 100);   
	}
	
	private Set<Address> createAddresses(int numNodes) 
	{      
		Set<Address> addresses = new HashSet<Address>(numNodes);      
		for (int i = 0; i < numNodes; i++) 
		{         
			addresses.add(new JGroupsAddress(org.jgroups.util.UUID.randomUUID()));      
		}      
		return addresses;   
	}   
	
	private ConsistentHash createConsistentHash(int numNodes, int virtualNodes) 
	{      
		Configuration c = new Configuration();      
		c.fluent().hash().consistentHashClass(DefaultConsistentHash.class).numVirtualNodes(virtualNodes);      
		//c.fluent().hash().consistentHashClass(ExperimentalDefaultConsistentHash.class).numVirtualNodes(numNodes*4);      
		Set<Address> addresses = createAddresses(numNodes);      
		return ConsistentHashHelper.createConsistentHash(c, addresses);   
	}   
	private List<PnLDataKey> buildKeys(int numKeys) {
		String[] legalEntities = {"F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8"};
		String[] transits = {"F1_1135", "F2_1135", "F3_1135", "F4_1135", "F5_1135", "F6_1135", "F7_1135", "F8_1135"};
		String[] books = {"F1_1135_NYAGENCIES", "F1_1135_NYDERIV", "F2_1135_NYAGENCIES", "F2_1135_NYDERIV", "F3_1135_NYAGENCIES", 
		"F3_1135_NYDERIV", "F4_1135_NYAGENCIES", "F4_1135_NYDERIV", "F5_1135_NYAGENCIES", "F5_1135_NYDERIV", "F6_1135_NYAGENCIES", 
		"F6_1135_NYDERIV", "F7_1135_NYAGENCIES", "F7_1135_NYDERIV", "F8_1135_NYAGENCIES", "F8_1135_NYDERIV"};
		String[] tradeIds = {"F1_1135_NYAGENCIES_0_0-8:Duco", "F1_1135_NYDERIV_0_0-12:Duco", "F1_1135_NYDERIV_1_0-2:Duco", "F2_1135_NYAGENCIES_0_0:Duco","F2_1135_NYAGENCIES_0_0-7:Duco", "F2_1135_NYAGENCIES_0_0-11:Duco", "F2_1135_NYAGENCIES_0_0-12:Duco", "F2_1135_NYAGENCIES_1_0-4:Duco",  "F2_1135_NYDERIV_0_0-1:Duco",
				"F3_1135_NYAGENCIES_0_0:Duco", "F3_1135_NYAGENCIES_1_0-3:Duco", "F6_1135_NYAGENCIES_0_0-8:Duco", "F6_1135_NYAGENCIES_0_0-10:Duco", "F6_1135_NYAGENCIES_0_0-11:Duco", "F6_1135_NYAGENCIES_1_0-2:Duco", "F6_1135_NYAGENCIES_1_0-3:Duco", "F6_1135_NYDERIV_0_0-8:Duco",
				"F6_1135_NYDERIV_0_0-14:Duco", "F7_1135_NYAGENCIES_0_0-3:Duco", "F7_1135_NYAGENCIES_0_0-6:Duco", "F7_1135_NYAGENCIES_0_0-13:Duco", "F8_1135_NYAGENCIES_0_0-10:Duco", "F8_1135_NYAGENCIES_1_0:Duco", "F8_1135_NYAGENCIES_1_0-2:Duco", "F8_1135_NYDERIV_0_0-12:Duco", "F8_1135_NYDERIV_1_0-4:Duco",
				"F3_1135_NYDERIV_0_0-5:Duco", "F3_1135_NYDERIV_0_0-6:Duco", "F3_1135_NYDERIV_0_0-7:Duco", "F3_1135_NYDERIV_0_0-8:Duco", "F3_1135_NYDERIV_0_0-9:Duco", "F4_1135_NYAGENCIES_0_0-7:Duco", "F4_1135_NYDERIV_0_0-7:Duco", "F4_1135_NYDERIV_1_0-3:Duco", "F5_1135_NYAGENCIES_0_0-8:Duco", "F5_1135_NYAGENCIES_0_0-14:Duco",
				"F5_1135_NYDERIV_0_0-2:Duco", "F5_1135_NYDERIV_1_0-1:Duco", "F1_1135_NYAGENCIES_0_0-2:Duco", "F1_1135_NYAGENCIES_0_0-10:Duco", "F1_1135_NYDERIV_0_0-7:Duco", "F1_1135_NYDERIV_1_0-1:Duco", "F1_1135_NYDERIV_1_0-3:Duco", "F2_1135_NYAGENCIES_0_0-6:Duco", "F2_1135_NYAGENCIES_0_0-8:Duco",
				"F2_1135_NYDERIV_0_0-3:Duco", "F2_1135_NYDERIV_0_0-9:Duco", "F2_1135_NYDERIV_0_0-11:Duco","F3_1135_NYAGENCIES_0_0-2:Duco","F6_1135_NYAGENCIES_0_0-9:Duco","F6_1135_NYAGENCIES_0_0-13:Duco","F6_1135_NYAGENCIES_1_0:Duco","F6_1135_NYAGENCIES_1_0-1:Duco","F6_1135_NYDERIV_0_0:Duco",
				"F6_1135_NYDERIV_0_0-11:Duco","F6_1135_NYDERIV_1_0:Duco","F7_1135_NYAGENCIES_1_0-1:Duco","F7_1135_NYDERIV_0_0-2:Duco","F7_1135_NYDERIV_1_0:Duco","F7_1135_NYDERIV_1_0-2:Duco","F8_1135_NYAGENCIES_0_0-3:Duco","F8_1135_NYAGENCIES_0_0-4:Duco",
				"F8_1135_NYAGENCIES_0_0-12:Duco","F8_1135_NYDERIV_0_0:Duco","F8_1135_NYDERIV_0_0-1:Duco","F8_1135_NYDERIV_0_0-10:Duco","F3_1135_NYDERIV_0_0-13:Duco","F3_1135_NYDERIV_1_0-3:Duco","F4_1135_NYAGENCIES_0_0-1:Duco","F4_1135_NYAGENCIES_0_0-6:Duco","F4_1135_NYAGENCIES_0_0-10:Duco",
				"F4_1135_NYAGENCIES_0_0-11:Duco","F4_1135_NYAGENCIES_0_0-12:Duco","F4_1135_NYAGENCIES_1_0-4:Duco","F4_1135_NYDERIV_0_0-12:Duco","F4_1135_NYDERIV_1_0-4:Duco","F5_1135_NYAGENCIES_0_0-1:Duco","F5_1135_NYAGENCIES_0_0-3:Duco","F5_1135_NYAGENCIES_0_0-5:Duco","F5_1135_NYAGENCIES_0_0-12:Duco","F5_1135_NYAGENCIES_1_0:Duco",
				"F5_1135_NYDERIV_0_0-5:Duco","F5_1135_NYDERIV_0_0-9:Duco","F5_1135_NYDERIV_0_0-13:Duco","F5_1135_NYDERIV_1_0-4:Duco","F6_1135_NYAGENCIES_0_0:Duco","F6_1135_NYAGENCIES_0_0-1:Duco","F1_1135_NYAGENCIES_0_0-7:Duco","F1_1135_NYDERIV_0_0-5:Duco",
				"F1_1135_NYDERIV_1_0-4:Duco","F2_1135_NYAGENCIES_0_0-2:Duco","F2_1135_NYAGENCIES_1_0-2:Duco","F2_1135_NYDERIV_1_0:Duco","F3_1135_NYAGENCIES_0_0-5:Duco","F3_1135_NYAGENCIES_0_0-6:Duco","F3_1135_NYAGENCIES_1_0-1:Duco","F6_1135_NYAGENCIES_0_0-14:Duco","F6_1135_NYAGENCIES_1_0-4:Duco","F6_1135_NYDERIV_0_0-3:Duco","F6_1135_NYDERIV_0_0-12:Duco"};

		List<PnLDataKey> keys = new ArrayList<PnLDataKey>();
		
		Random rand = new Random();
		for(int i=0 ; i< numKeys; i++) {
			PnLDataKey key = new PnLDataKey(legalEntities[rand.nextInt(legalEntities.length)], 
					transits[rand.nextInt(transits.length)], 
					books[rand.nextInt(books.length)], 
					tradeIds[rand.nextInt(tradeIds.length)]);
			keys.add(key);
		}
		
		return keys;
	}
} 
